package web.run

import japgolly.scalajs.react.Callback
import monix.execution.Ack
import monix.reactive.{Observable, Observer}
import monix.reactive.observers.Subscriber
import org.scalajs.dom
import org.scalajs.dom.ext.{Ajax, AjaxException}
import relay.runtime.{network, Disposable, Handler}
import relay.runtime.RelayRuntime.{ConnectionHandler, ViewerHandler}
import relay.runtime.network.{FetchFunction, QueryPayload, SubscribeFunction}
import relay.runtime.store.{RelayEnvironment, RelayEnvironmentConfig}
import web.base.all._
import web.base.mnx._

import scala.concurrent.{Future, Promise}
import scala.scalajs.js
import scala.scalajs.js.Dynamic.literal
import scala.scalajs.js.{JSON, JavaScriptException}
import scala.scalajs.js.JSConverters._
import scala.util.{Failure, Success}
import scala.concurrent.duration._

object Env {

  def handlerProvider(n: String): Handler = n match {
    case "viewer"     => ViewerHandler
    case "connection" => ConnectionHandler
    case _            => throw new Exception("No handler")
  }

  /**
    * Similar to the SimpleNetwork from classic.
    */
  val fetch: FetchFunction = { (operation, variables, cacheConfig, uploadables) =>
    val headers = Map("Accept" -> "*/*", "X-Requested-With" -> "Relay", "Content-Type" -> "application/json")

    val p = Promise[QueryPayload]()

    Ajax
      .post(
        "/gql",
        JSON.stringify(literal(query = operation.text, variables = variables)),
        headers         = headers,
        responseType    = "json",
        withCredentials = true)
      .onComplete {
        case Success(f) =>
          p.success(f.response.asInstanceOf[QueryPayload])

        case Failure(AjaxException(f)) if f.status == 400 =>
          p.failure(JavaScriptException(f.response.asInstanceOf[QueryPayload]))

        case Failure(AjaxException(f)) if f.status == 401 || f.status == 403 =>
          p.failure(JavaScriptException(f.response.asInstanceOf[QueryPayload]))

        case Failure(aje @ AjaxException(_)) =>
          p.failure(aje)

        case Failure(f) =>
          p.failure(js.JavaScriptException(f))
      }

    p.future.toJSPromise
  }

  /**
    * Singleton websocket a multiplexed capable websocket client.
    */
  lazy val websocket = new WsClient("event-stream")

  trait SeedInterface extends js.Object {
    val seed: js.UndefOr[String]
  }

  /**
    * Function to handle subscribing.
    */
  val subFn: SubscribeFunction = { (batch, variables, cacheConfig, relayObserver) =>
    val seedDefault = System.currentTimeMillis().toString
    val seed        = variables.asInstanceOf[SeedInterface].seed.getOrElse(seedDefault)
    val vars        = variables.asInstanceOf[js.Dynamic]
    vars.seed = seed
    dom.console.log(s"Subscribe function called on seed $seed.", vars)

//    val cancelable =
//      websocket.submitQuery(seed = seed,
//                            query = batch.text,
//                            variables = vars.asInstanceOf[js.Object],
//                            operation = batch.name)(
//        new Observer[SubEventFrame] {
//          def onError(ex: Throwable) = relayObserver.onError.foreach { oe =>
//            oe(new js.Error(ex.getMessage))
//          }
//
//          def onComplete() = relayObserver.onCompleted.foreach(_())
//
//          def onNext(elem: SubEventFrame) = {
//            relayObserver.onNext.foreach(_(elem.asInstanceOf[QueryPayload]))
//            Ack.Continue
//          }
//        }
//      )
//
//    new Disposable {
//      def dispose() = {
//        cancelable.cancel()
//      }
//    }

    val query = WebSocketQuery.submit(seed, batch.text, batch.name, vars.asInstanceOf[js.Object])
    val channel = websocket
      .doOnSubscribe { () =>
        websocket.onNext(query); ()
      }
      .filter(_.seed == seed)
      .subscribe(new Observer[EventFrameMsg] {
        override def onError(ex: Throwable): Unit = relayObserver.onError.foreach { oe =>
          oe(new js.Error(ex.getMessage))
        }

        override def onComplete(): Unit = relayObserver.onCompleted.foreach(_())

        override def onNext(elem: EventFrameMsg): Future[Ack] = {
          relayObserver.onNext.foreach(_(elem.asInstanceOf[QueryPayload]))
          Ack.Continue
        }
      })

    new Disposable {
      override def dispose(): Unit = channel.cancel()
    }
  }

  def makeEnvironment = {
    Callback.log(s"$toString Created environment").runNow()
    val source      = new relay.runtime.store.RelayRecordSource()
    val store       = new relay.runtime.store.RelayStore(source)
    val net         = network.RelayNetwork.create(fetch, subFn)
    val environment = new RelayEnvironment(RelayEnvironmentConfig(store, net, js.defined(handlerProvider)))
    environment
  }

  override def toString = "[Env]"
}

@js.native
trait EventFrame extends QueryPayload {
  val event: js.UndefOr[String] = js.undefined
}
