package web.admin.run

import japgolly.scalajs.react.Callback
import monix.execution.Ack
import monix.reactive.Observer
import org.scalajs.dom
import org.scalajs.dom.ext.{Ajax, AjaxException}
import relay.runtime.{network, Disposable}
import relay.runtime.network.{FetchFunction, QueryPayload, SubscribeFunction}
import relay.runtime.store.{RelayEnvironment, RelayEnvironmentConfig}
import web.base.all._
import web.base.mnx.{SubEventFrame, WebSocketClient}

import scala.concurrent.Promise
import scala.scalajs.js
import scala.scalajs.js.Dynamic.literal
import scala.scalajs.js.{JSON, JavaScriptException}
import scala.scalajs.js.JSConverters._
import scala.util.{Failure, Success}

object Env {

  /**
    * Similar to the SimpleNetwork from classic.
    */
  val fetch: FetchFunction = { (operation, variables, cacheConfig, uploadables) =>
    val headers = Map("Accept" -> "*/*", "X-Requested-With" -> "Relay", "Content-Type" -> "application/json")

    val p = Promise[QueryPayload]()

    Ajax
      .post(
        "/agql",
        JSON.stringify(literal(query = operation.text, variables = variables)),
        headers         = headers,
        responseType    = "json",
        withCredentials = true)
      .onComplete {
        case Success(f) =>
          p.success(f.response.asInstanceOf[QueryPayload])

        case Failure(AjaxException(f)) if f.status == 400 =>
          p.failure(JavaScriptException(f.response.asInstanceOf[QueryPayload]))

        case Failure(AjaxException(f)) if f.status == 401 || f.status == 403 =>
          p.failure(JavaScriptException(f.response.asInstanceOf[QueryPayload]))

        case Failure(aje @ AjaxException(_)) =>
          p.failure(aje)

        case Failure(f) =>
          p.failure(js.JavaScriptException(f))
      }

    p.future.toJSPromise
  }

  /**
    * Singleton websocket a multiplexed capable websocket client.
    */
  lazy val websocket = new WebSocketClient("event-stream-admin")

  /**
    * Function to handle subscribing.
    */
  val subFn: SubscribeFunction = { (batch, variables, cacheConfig, relayObserver) =>
    dom.console.log("Subscribe function called.")
    val seed = variables
      .asInstanceOf[js.Dynamic]
      .seed
      .asInstanceOf[js.UndefOr[String]]
      .getOrElse(System.currentTimeMillis().toString)
    //

    val cancelable =
      websocket.submitQuery(seed = seed, query = batch.text, variables = variables, operation = batch.name)(
        new Observer[SubEventFrame] {
          def onError(ex: Throwable) = relayObserver.onError.foreach { oe =>
            oe(new js.Error(ex.getMessage))
          }

          def onComplete() = relayObserver.onCompleted.foreach(_())

          def onNext(elem: SubEventFrame) = {
            relayObserver.onNext.foreach(_(elem.asInstanceOf[QueryPayload]))
            Ack.Continue
          }
        })

    new Disposable {
      def dispose() = {
        cancelable.cancel()
      }
    }
  }

  /**
    * Environment factory.
    * @return
    */
  def makeEnvironment = {
    Callback.log(s"$toString Created environment").runNow()
    val source      = new relay.runtime.store.RelayRecordSource()
    val store       = new relay.runtime.store.RelayStore(source)
    val net         = network.RelayNetwork.create(fetch, subFn)
    val environment = new RelayEnvironment(RelayEnvironmentConfig(store, net))
    environment
  }

  override def toString = "[AdminEnv]"
}
